﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Operators {
  [Item("IntDataReducer", "An operator to reduce int values of sub scopes")]
  [StorableClass]
  public sealed class IntDataReducer : SingleSuccessorOperator {
    #region Parameter Properties
    public ScopeTreeLookupParameter<IntValue> ParameterToReduce {
      get { return (ScopeTreeLookupParameter<IntValue>)Parameters["ParameterToReduce"]; }
    }
    public LookupParameter<IntValue> TargetParameter {
      get { return (LookupParameter<IntValue>)Parameters["TargetParameter"]; }
    }
    public ValueParameter<ReductionOperation> ReductionOperation {
      get { return (ValueParameter<ReductionOperation>)Parameters["ReductionOperation"]; }
    }
    public ValueParameter<ReductionOperation> TargetOperation {
      get { return (ValueParameter<ReductionOperation>)Parameters["TargetOperation"]; }
    }
    #endregion

    [StorableConstructor]
    private IntDataReducer(bool deserializing) : base(deserializing) { }
    private IntDataReducer(IntDataReducer original, Cloner cloner)
      : base(original, cloner) {
    }
    public IntDataReducer()
      : base() {
      #region Create parameters
      Parameters.Add(new ScopeTreeLookupParameter<IntValue>("ParameterToReduce", "The parameter on which the reduction operation should be applied."));
      Parameters.Add(new LookupParameter<IntValue>("TargetParameter", "The target variable in which the reduced value should be stored."));
      Parameters.Add(new ValueParameter<ReductionOperation>("ReductionOperation", "The operation which is applied on the parameter to reduce."));
      Parameters.Add(new ValueParameter<ReductionOperation>("TargetOperation", "The operation used to apply the reduced value to the target variable."));
      #endregion
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new IntDataReducer(this, cloner);
    }

    public override IOperation Apply() {
      var intValues = ParameterToReduce.ActualValue;

      if (intValues.Count() > 0) {
        if (TargetParameter.ActualValue == null) TargetParameter.ActualValue = new IntValue();

        int result = 1;
        switch (ReductionOperation.Value.Value) {
          case ReductionOperations.Sum:
            result = intValues.Sum(x => x.Value);
            break;
          case ReductionOperations.Prod:
            intValues.ForEach(x => result *= x.Value);
            break;
          case ReductionOperations.Avg:
            result = (int)Math.Round(intValues.Average(x => x.Value));
            break;
          case ReductionOperations.Min:
            result = intValues.Min(x => x.Value);
            break;
          case ReductionOperations.Max:
            result = intValues.Max(x => x.Value);
            break;
          default:
            throw new InvalidOperationException("Operation " + ReductionOperation.Value.Value + " is not supported as ReductionOperation.");
        }

        switch (TargetOperation.Value.Value) {
          case ReductionOperations.Assign:
            TargetParameter.ActualValue.Value = result;
            break;
          case ReductionOperations.Sum:
            TargetParameter.ActualValue.Value += result;
            break;
          case ReductionOperations.Prod:
            if (TargetParameter.ActualValue.Value == 0) TargetParameter.ActualValue.Value = 1;
            TargetParameter.ActualValue.Value *= result;
            break;
          default:
            throw new InvalidOperationException("Operation " + TargetOperation.Value.Value + " is not supported as TargetOperation.");
        }
      }
      return base.Apply();
    }
  }
}
